"""A set of convenience functions to search in fields"""
import re
from typing import Generator

from unidecode import unidecode
from tinydb import database

from . import model


def creator_name(corpusdb: database.TinyDB,
                 pattern: str) -> Generator[model.Book, None, None]:
    """Return a list of document based on creator's name. It completely
    ignores accents and case."""
    for document in iter(corpusdb):
        if any(re.search(unidecode(pattern),
                        unidecode(creator['name']),
                        flags=re.IGNORECASE)
               for creator in document['creator']):
            yield model.Book(**document)


def title(corpusdb: database.TinyDB,
          pattern: str) -> Generator[model.Book, None, None]:
    """Return a list of documents based on name. It completely
    ignores accents and case"""
    for document in iter(corpusdb):
        if any(re.search(unidecode(pattern),
                        unidecode(document_title),
                        flags=re.IGNORECASE)
               for document_title in document['title']):
            yield model.Book(**document)



"""Define the litcorpt data models"""
# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-self-use
# pylint: disable=too-few-public-methods

from enum import Enum
from typing import List, Optional
from datetime import date
from pydantic import BaseModel, ValidationError


# class ISBN10FormatError(Exception):
#     """Custom error that is raised when ISBN10 doesn't has the right format"""
#
#     def __init__(self, value: str, message) -> None:
#         self.value = value
#         self.message = message
#         super().__init__(message)


# class IdentifierMissingError(Exception):
#     """Custom error that is raised when there is no identifier"""
#
#     def __init__(self, title: str, message) -> None:
#         self.title = title
#         self.message = message
#         super().__init__(message)

##############################################################
# Enumeration classes
##############################################################
class CreatorRolesEnum(str, Enum):
    """Enumeration for Creator Roles"""
    AUTHOR: str = 'author'
    EDITOR: str = 'editor'
    TRANSLATOR: str = 'translator'
    ORGANIZER: str = 'organizer'


class OrtographyEnum(str, Enum):
    """Enumeration for Ortographic Agreements
       https://pt.wikipedia.org/wiki/Ortografia_da_l%C3%ADngua_portuguesa
    """
    FN1200: str = '1200fonetico'
    ET1500: str = '1500etimologico'
    PT1885: str = '1885portugal'
    BR1907: str = '1907brasil'
    PT1911: str = '1911portugal'
    BR1915: str = '1915brasil'
    LS1931: str = '1931lusofono'
    BR1943: str = '1943brasil'
    LS1945: str = '1945lusofono'
    BR1971: str = '1971brasil'
    PT1973: str = '1973portugal'
    LS1975: str = '1975lusofono'
    BR1986: str = '1986brasil'
    LS1990: str = '1990lusofono'

class LicenseEnum(str, Enum):
    """License Enumeration"""
    PUBLIC: str = 'Public Domain'
    PRIVATE: str = 'Copyrighted'
    CC: str = 'Creative Commons'
    ACADEMIC: str = 'Free for academic usage'
    SPECIAL: str = 'Special use'
    UNKNOWN: str = 'Unknown'


##############################################################
# Model classes
##############################################################
class Creator(BaseModel):
    """Define creator elements"""
    role: CreatorRolesEnum
    name: Optional[str]
    birth: Optional[int]
    death: Optional[int]
    place: Optional[str]

class Identifiers(BaseModel):
    """Define creator elements"""
    isbn_10: Optional[str]
    isbn_13: Optional[str]
    amazon_books: Optional[str]
    google_books: Optional[str]
    goodreads: Optional[str]
    doi: Optional[str]


class Book(BaseModel):
    """Book entry model

    This class define the elements of a book entry.

    index:             An unique string to internaly identify the entry.
    title:             A title associated to the entry.
    subtitle:          Document subtitle (if exists)
    creator:           A list of creators. Each creator contains:
                           Role: Creator relationship with the book entry
                           LastName: creator last name, often used in bibliography,
                           FirstName: creator given name,
                           Birth: Creator's birth year.
                           Death: Creator's death year.
                           Place: Creator's birth place.
    language:          A list of ISO entry with language, pt_BR or pt are the
                       most common here.  A document can contain many languages.
                       Most of time just one.
    published:         Date of first publish. Multiple edition should use the
                       date of first edition. Except when large changes happened
                       in document, as change of translator, change of ortography.
    identifier:        A unique global identifier, often a ISBN13 for books.
    original_language: Original language of document. Using ISO entry for language.
    subject:           A list entry subjects. As example: Fantasy, Science-Fiction, Kids.
                       Use lower caps always.
    genre:             A list of literary genre: Novel, Poetry, Lyrics, Theather.
                       Use lower caps always.
    ortography:        Reference to which Portuguese ortography is being used.
    abstract:          The book abstract/resume.
    notes:             Any notes worth of note.
    contents:          Book contents. The text itself.

    Caution: Date fields must contain a datetime.date or a string in format "YYYY-MM-DD"
    """

    index: str
    title: str
    subtitle: Optional[str]
    creator: List[Creator]
    language: List[str]
    published: Optional[date]
    identifier: Optional[List[str]]
    identifiers: Optional[List[Identifiers]]
    original_language: Optional[List[str]]
    subject: Optional[List[str]]
    genre: Optional[List[str]]
    ortography: Optional[OrtographyEnum]
    abstract: Optional[str]
    notes: Optional[List[str]]
    license: Optional[LicenseEnum] = LicenseEnum.UNKNOWN
    contents: Optional[str]

    # @validator('subject')
    # def validate_list_lower_subject(cls, field: Optional[str]) -> None:
    #     """Every entry must be in lower case"""
    #     if field is not None:
    #         for entry in field:
    #             if entry != entry.lower():
    #                 raise ValueError(f"""Entry '{entry}' must be lower case""")

    # @validator('genre')
    # def validate_list_lower_genre(cls, field: Optional[str]) -> None:
    #     """Every entry must be in lower case"""
    #     if field is not None:
    #         for entry in field:
    #             if entry != entry.lower():
    #                 raise ValueError(f"""Entry '{entry}' must be lower case""")

    # @root_validator(pre=True)
    # @classmethod
    # def check_isbn10_or_isbn13(cls, values):
    #     """Makes sure that isbn10 or isbn13 are present"""
    #     if "isbn_10" not in values and "isbn_13" not in values:
    #         raise IdentifierMissingError(title=values['title'],
    #                                      message="Document should contain a identifier")
    #     return values


    # @validator('isbn_10')
    # @classmethod
    # def isbn_10_valid(cls, value):
    #     chars = [c for c in value if c in "0123456789Xx"]
    #     if len(chars != 10):
    #         raise ISBN10FormatError(value=value, message="ISBN10 should be 10 digits.")

    #     def char_to_int(char: str)-> int:
    #         if char in "Xx":
    #             return 10
    #         return int(char)

    #     weighted_sum  = sum((10 -i ) * char_to_int(x) for u, x in enumerate(chars))
    #     if weighted_sum % 11 != 0:
    #         raise ISBN10FormatError(value=value, message="ISBN10 should be divisible by 11."
    #     return value

##############################################################
# Main test function
##############################################################
def test_book() -> Book:
    """Return a Book for testing purposes."""
    assis_memorias = {
        'index': 'assismemorias1880',
        'title': [ 'Memórias Póstumas de Brás Cubas', 'Epitáfio de um Pequeno Vencedor' ],
        'creator': [
            {'role': 'author',
             'name': 'Machado de Assis',
             'birth': '1839',
             'death': '1908',
            }
        ],
        'language': ['pt_BR'],
        'published': '1881-01-01',
        'original_language': ['pt_BR'],
        'subject': ['realism', 'humor'],
        'genre': ['novel', 'prose'],
        'ortography': '1990lusofono'
    }

    return Book(**assis_memorias)


#%%
def test_book_contents() -> str:
    """Return a book contents for testing"""
    memorias = """
                                        AO VERME
                                           QUE
                              PRIMEIRO ROEU AS FRIAS CARNES
                                      DO MEU CADÁVER
                                          DEDICO
                                  COMO SAUDOSA LEMBRANÇA
                                          ESTAS

                                    Memórias Póstumas



Prólogo da quarta edição


A PRIMEIRA edição destas Memórias póstumas de Brás Cubas foi feita aos
pedaços na Revista Brasileira, pelos anos de 1870. Postas mais tarde em
livro, corrigi o texto em vários logares. Agora que tive de o rever para
a terceira edição, emendei ainda alguma cousa e suprimi duas ou três
dúzias de linhas. Assim composto, sai novamente à luz esta obra que
alguma benevolência parece ter encontrado no público.
Capistrano de Abreu, noticiando a publicação do livro, perguntava:
« As Memórias Póstumas de Brás Cubas são um romance? » Macedo So-
ares, em carta que me escreveu por esse tempo, recordava amigamente
as Viagens na minha terra. Ao primeiro respondia já o defuncto Brás Cu-
bas (como o leitor viu e verá no prólogo dele que vai adeante) que sim
e que não, que era romance para uns e não o era para outros. Quanto
ao segundo, assim se explicou o finado: « Trata-se de uma obra difusa,
na qual eu, Brás Cubas, se adoptei a forma livre de um Sterne ou de
um Xavier de Maistre, não sei se lhe meti algumas rabugens de pessi-
mismo. » Toda essa gente viajou: Xavier de Maistre à roda do quarto,
Garrett na terra dele, Sterne na terra dos outros. De Brás Cubas se pode
talvez dizer que viajou à roda da vida.
O que faz do meu Brás Cubas um autor particular é o que ele chama
« rabugens de pessimismo » . Há na alma deste livro, por mais risonho
que pareça, um sentimento amargo e áspero, que está longe de vir dos
seus modelos. É taça que pode ter lavores de igual escola, mas leva outro
vinho. Não digo mais para não entrar na crítica de um defunto, que se
pintou a si e a outros, conforme lhe pareceu melhor e mais certo.

MACHADO DE A SSIS


Ao leitor

QUE STENDHAL confessasse haver escripto um de seus livros para cem
leitores, cousa é que admira e consterna. O que não admira, nem pro-
vavelmente consternará é se este outro livro não tiver os cem leitores de
Stendhal, nem cincoenta, nem vinte, e quando muito, dez. Dez? Talvez
cinco. Trata-se, na verdade, de uma obra difusa, na qual eu, Brás Cu-
bas, se adoptei a forma livre de um Sterne, ou de um Xavier de Maistre,
não sei se lhe meti algumas rabugens de pessimismo. Pode ser. Obra de
finado. Escrevi-a com a pena da galhofa e a tinta da melancolia, e não
é difícil antever o que poderá sair desse conúbio. Acresce que a gente
grave achará no livro umas aparências de puro romance, ao passo que a
gente frívola não achará nele o seu romance usual; ei-lo aí fica privado
da estima dos graves e do amor dos frívolos, que são as duas colunas
máximas da opinião.
Mas eu ainda espero angariar as simpatias da opinião, e o primeiro
remédio é fugir a um prólogo explícito e longo. O melhor prólogo é o que
contém menos cousas, ou o que as diz de um jeito obscuro e truncado.
Conseguintemente, evito contar o processo extraordinário que empre-
guei na composição destas Memórias, trabalhadas cá no outro mundo.
Seria curioso, mas nimiamente extenso, e aliás desnecessário ao enten-
dimento da obra. A obra em si mesma é tudo: se te agradar, fino leitor,
pago-me da tarefa; se te não agradar, pago-te com um piparote, e adeus.
BRÁS CUBAS


CAPÍTULO PRIMEIRO

Óbito do autor
ALGUM tempo hesitei se devia abrir estas memórias pelo princípio ou
pelo fim, isto é, se poria em primeiro logar o meu nascimento ou a
minha morte. Suposto o uso vulgar seja começar pelo nascimento, duas
considerações me levaram a adoptar diferente método: a primeira é que
eu não sou propriamente um autor defunto, mas um defunto autor, para
quem a campa foi outro berço; a segunda é que o escripto ficaria assim
mais galante e mais novo. Moisés, que também contou a sua morte, não
a pôs no intróito, mas no cabo: diferença radical entre este livro e o
Pentateuco.

Continua...
"""

    return memorias


def main() -> None:
    """Run basic tests for models"""
    try:
        assis_book = test_book()
    except ValidationError as err:
        print(f"Error in book: {err}")
    else:
        print(f"Python Object:\n {assis_book}\n")
        print(f"Json   Object:\n {assis_book.json()}\n")


if __name__ == "__main__":
    main()



"""litcorpt module initialization"""

from tinydb import Query


from .constants import (__version__,
                        __author__,
                        __author_email__,
)

from .main import (corpus,
                   corpus_load,
                   corpus_read,
                   corpus_write,
                   corpus_retrieve,
                   corpus_build_database,
                   corpus_insert_book,
                   book_write,
                   book_read,
                   metadata,
                   doc_id,
                   logger,
)

from .search import (creator_name,
                     title)
from .epub import read



"""This module contains utilities for litcorpt API"""

import os
import re
import logging

from itertools import cycle
from typing import Union, Generator, List

from urllib.parse import urlparse

from pathlib import Path
from dotenv import load_dotenv

import requests

logger: logging.Logger = logging.getLogger(__name__)

def spinning_wheel() -> Generator[None, None, None]:
    """Print a frame of a spinning wheel"""
    #whell: str=r"-\|/"
    #whell: str=r"⣾⣽⣻⢿⡿⣟⣯⣷"
    #whell: str=r"🕛🕐🕑🕒🕓🕔🕕🕖🕗🕘🕙🕚"
    # whell: List[str]= ["▹▹▹▹▹",
    #                    "▸▹▹▹▹",
    #                    "▹▸▹▹▹",
    #                    "▹▹▸▹▹",
    #                    "▹▹▹▸▹",
    #                    "▹▹▹▹▸"]
    whell: List[str] = ["⢀⠀", "⡀⠀", "⠄⠀", "⢂⠀", "⡂⠀", "⠅⠀", "⢃⠀", "⡃⠀", "⠍⠀",
                        "⢋⠀", "⡋⠀", "⠍⠁", "⢋⠁", "⡋⠁", "⠍⠉", "⠋⠉", "⠋⠉", "⠉⠙",
                        "⠉⠙", "⠉⠩", "⠈⢙", "⠈⡙", "⢈⠩", "⡀⢙", "⠄⡙", "⢂⠩", "⡂⢘",
                        "⠅⡘", "⢃⠨", "⡃⢐", "⠍⡐", "⢋⠠", "⡋⢀", "⠍⡁", "⢋⠁", "⡋⠁",
                        "⠍⠉", "⠋⠉", "⠋⠉", "⠉⠙", "⠉⠙", "⠉⠩", "⠈⢙", "⠈⡙", "⠈⠩",
                        "⠀⢙", "⠀⡙", "⠀⠩", "⠀⢘", "⠀⡘", "⠀⠨", "⠀⢐", "⠀⡐", "⠀⠠",
                        "⠀⢀", "⠀⡀"]
    for frame in cycle(whell):
        print(f'{frame}\r', sep='', end='', flush=True)
        yield


def download_file(url: str, dstdir: str = '.', filename: str = None):
    """Download file from url and save as dstdir/filepath. If filename is omitted
       check if http requests gives a filename, otherwise use url path
       filename and save at dstdir. This do not create the
       dir structure, that must be created beforehand.

    Input:
        url (str): Url to download
        filename (str): Filename to download to. If missing try to figure out.
        dstdir (str): Directory to store data

    Outputs:
        filepath (str): Filepath to downloaded file
        filesize (int): Ammount of bytes downloaded
    """

    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        # Check if filepath was given, otherwise find from request
        if filename is None:
            # Name from http headers
            if 'content-disposition' in req.headers:
                disposition: str = req.headers['content-disposition']
                filename = re.findall("filepath=(.+)", disposition)[0]

            # We got a redirect
            if req.history:
                hist_req: requests.models.Response = req.history[0]
                if hist_req.status_code == 302:
                    url = hist_req.headers.get('Location', url)

            # Name from url path
            if not filename:
                filename = urlparse(url).path.split('/')[-1]

        # Check filesize (only works when stream=False)
        expected_filesize: int = 0
        if 'Content-Length' in req.headers:
            expected_filesize = int(req.headers['Content-Length'])

        filepath = os.path.join(dstdir, filename)
        logger.info('Downloading %s to %s', url, filepath)
        if expected_filesize > 0:
            logger.info('Expected content size %s bytes', expected_filesize)

        filesize = 0
        with open(filepath, 'wb') as f:
            spinning_wheel_frame = spinning_wheel()
            for chunk in req.iter_content(chunk_size=None):
                if chunk:
                    next(spinning_wheel_frame)
                    filesize += len(chunk)
                    f.write(chunk)

    return filepath, filesize


def get_corpus_datapath() -> Path:
    """Returns a Path to CORPUS_DATAPATH"""
    load_dotenv()
    corpus_datapath_default: str = '~/litcorpt_data'
    corpus_datapath: Path = Path(os.getenv('CORPUS_DATAPATH',
                                           corpus_datapath_default))
    corpus_datapath = corpus_datapath.expanduser()
    corpus_datapath.mkdir(parents=True, exist_ok=True)

    return corpus_datapath


def get_corpus_url() -> str:
    """Returns a URL string to CORPUS source """
    # pylint: disable=line-too-long
    load_dotenv()
    corpus_url_default: str = "https://github.com/igormorgado/litcorpt_data/archive/refs/heads/main.zip"
    corpus_url = os.getenv('CORPUS_URL', corpus_url_default)

    return corpus_url


def book_dir(path: Union[str, Path]) -> Path:
    """ Return a bookdir path, given a datapath. Often datapath is
    read from CORPUS_DATAPATH environment variable"""
    # Resolve the zip file path.  zip dirname is: repository-branch
    datadir: str = 'litcorpt_data-main/data/'
    booksdir: Path = Path(path) / datadir
    return booksdir



#!/usr/bin/env python
"""Entry point for Literary corpus in Portuguese"""

import shutil
import tempfile
import logging
from  datetime import datetime

from typing import List, Union, Optional, Dict, Any
from pathlib import Path
from tinydb import TinyDB, database, table, queries, Query

from . import utils
from . import model
from . import constants


############################################################
# LOGGER SETTINGS
############################################################
logger: logging.Logger = logging.getLogger(__name__)

# Log file handler
logger_filename: str = 'litcorpt_debug.log'
logger_filepath: Path = utils.get_corpus_datapath() / logger_filename
logger_fd: logging.FileHandler = logging.FileHandler(logger_filepath)

# Log console handler
logger_console: logging.StreamHandler = logging.StreamHandler()

# Log formatting
# pylint: disable=line-too-long
formatstr: str = "%(asctime)s:%(module)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s"
logger_fmt: logging.Formatter = logging.Formatter(formatstr)

# Log settings
# INFO, WARN, ERROR, CRITICAL go to file
logger_fd.setFormatter(logger_fmt)
logger_fd.setLevel(logging.INFO)

# WARN, ERROR and CRITICAL go to console
logger_console.setFormatter(logger_fmt)
logger_console.setLevel(logging.WARN)

# Default logger
logger.setLevel(logging.INFO)
logger.addHandler(logger_console)
logger.addHandler(logger_fd)


def book_json_path(path: Union[str, Path]) -> Path:
    """Return a book's json path, from a dirpath

    Input:
        path: A dir path containing a book.

    Outputs:
        path: A file path pointing to  book json file (if exists).
    """
    path = Path(path)
    basename: str = path.name
    json_filepath: Path = path / f"{basename}.json"
    return json_filepath


def book_contents_path(path: Union[str, Path]) -> Path:
    """Return a book's json path, from a dirpath

    Input:
        path: A dir path containing a book.

    Outputs:
        path: A file path pointing to  book contents file (if exists).
    """
    path = Path(path)
    basename: str = path.name
    text_filepath: Path = path / f"{basename}.txt"
    return text_filepath


def book_read_json(path: Union[str, Path]) -> model.Book:
    """Read a book json file and returns as a Book """
    return model.Book.parse_file(path)


def book_read_contents(path: Union[str, Path]) -> str:
    """Read book contents from filepath"""
    with open(path, 'r') as contents_fd:
        contents: str = contents_fd.read()
    return contents


def book_read(path: Union[str, Path]) -> model.Book:
    """Read a book json and contents from file path and returns
    as a Book object. If a content text file exists at same
    dir, loads into Book.contents

    Input:
        path: A full path to book json file, or to dirpath where json is.

    Outputs:
        model.Book: A Book object
    """
    path = Path(path)
    basedir: Path
    json_filepath: Path
    text_filepath: Path

    if path.is_file():
        # We have a json filepath.
        basedir  = path.parent
        json_filepath = path
        text_filepath = book_contents_path(basedir)
    else:
        # We probably received a basedir
        basedir = path
        json_filepath = book_json_path(path)
        text_filepath = book_contents_path(path)

    # Righ now we have all information we need to read the data
    book: model.Book = book_read_json(json_filepath)
    if book.contents is None and text_filepath.exists():
        contents: str = book_read_contents(text_filepath)
        book.contents = contents

    return book


def book_write_contents(contents: Union[str, model.Book],
                        path: Path) -> None:
    """Writes book contents into filepath"""

    local_contents: Optional[str]

    if isinstance(contents, model.Book):
        local_contents = contents.contents
    elif isinstance(contents, str):
        local_contents = contents
    else:
        errmsg: str = f"contents: Expected Union[str, model.Book], but got {type(contents)}"
        raise TypeError(errmsg)

    with open(path, 'w') as contents_fd:
        contents_fd.write(str(local_contents))


def book_write_json(book: model.Book,
                    path: Union[str, Path],
                    with_contents=False) -> None:
    """Writes book json into filepath.

    Inputs:
        book: A book object

    Outputs:
    """
    json_args: Dict = {'indent': 2, 'ensure_ascii': True}
    model_args: Dict = {'exclude_none': True}
    book_json: str

    if not with_contents:
        book_json = book.json(exclude={'contents'}, **model_args, **json_args)
    else:
        book_json = book.json(indent=2, **model_args, **json_args)
    with open(path, 'w') as book_fp:
        book_fp.write(book_json)


def book_write(book: model.Book, path: Union[Path, str]) -> None:
    """Writes the Book object in a book dirpath. If book model contains
    the ebook content. It will write the json WITHOUT the contents and
    the contents into a different txt file at the same directory.

    Filenames will be fetched from latest directory in Path.

    For example:

    `book_write(book, '~/data/books/book123/')`

    It will create:

    ```
    ~/data/books/book123/book123.json
    ~/data/books/book123/book123.txt
    ```

    Directory '~/data/books/book123/' is created if does not exist, but
    if dirpath is a file. It raises an exception NotADirectoryError

    Inputs:
        book: The book to be written
        path: Location to write to.
    """
    # Ensures dirpath is a Path and expanded
    local_path: Path = Path(path).expanduser()

    if not local_path.exists():
        Path(local_path).mkdir(parents=True)
    elif not local_path.is_dir():
        raise NotADirectoryError(f"{path}")

    basename: str = local_path.name

    json_filepath: Path = local_path / f"{basename}.json"
    book_write_json(book=book, path=json_filepath, with_contents=False)

    text_filepath: Path = local_path / f"{basename}.txt"
    book_write_contents(contents=book, path=text_filepath)


def corpus_retrieve(corpus_datapath: Union[str, Path], corpus_url: str) -> None:
    """This function assures that corpus is available in CORPUS_DATAPATH.
    If not it will download it.
    """
    corpus_datapath = Path(corpus_datapath)
    checkin_file: Path = corpus_datapath / 'last.txt'
    logger.debug("Reading checkin file %s", checkin_file)
    if not checkin_file.exists():
        logger.info("START: Corpus retrieval.")
        logger.warning("Corpus not found in %s.", corpus_datapath)

        corpus_datapath.mkdir(parents=True, exist_ok=True)
        with tempfile.TemporaryDirectory() as tmpdirname:
            # Build destination filepath
            filename: str = corpus_url.split('/')[-1]

            logger.warning("Downloading from %s.", corpus_url)
            logger.debug("Download destination %s/%s", tmpdirname, filename)
            # Download corpus
            filepath, _ = utils.download_file(corpus_url,
                                              dstdir=tmpdirname,
                                              filename=filename)

            logger.debug("Download finished %s", filepath)

            logger.info("Extracting zip to %s", corpus_datapath)

            # Extract Corpus
            shutil.unpack_archive(filepath, corpus_datapath, "zip")

        # Fill the success timestamp
        with open(checkin_file, 'w') as checkin_fd:
            now: datetime = datetime.now()
            timestamp: str = f'{now.year}{now.month}{now.day}{now.hour}{now.minute}{now.second}'
            checkin_fd.write(timestamp)

        logger.info("END: Corpus retrieval.")
    else:
        logger.info("Corpus found in %s.", corpus_datapath)


def corpus_read(path: Path) -> List[model.Book]:
    """Load corpus from corpus_path and return a list of Books

    The corpus_path must containg directories representing books
    as defined in book_read() function.
    """
    books: List[model.Book] = []
    for bookdir in path.iterdir():
        books.append(book_read(bookdir))

    return books


def corpus_write(corpus_data: List[model.Book],
                 path: Union[str, Path],
                 overwrite=False) -> List[bool]:
    """Write the whole corpus (a list of Books) into corpus_path.

    Inputs:
        corpus: A corpus structured as a List of model.Book objects.
        corpus_path: The dir path to corpus
        overwrite: Overwrite data if already exists.

    Outputs:
        status: A list with success status, related to each element in
                corpus input.
    """
    path = Path(path)
    write_status: List[bool] = []
    for book in corpus_data:
        bookpath: Path = path / book.index
        if bookpath.exists():
            if overwrite:
                # We need to catch some rmtree, add to failed status
                # and continue.
                shutil.rmtree(bookpath, ignore_errors=False, onerror=None)
            else:
                logger.warning("Book %s. Already exists. Skipping", book.index)
                write_status.append(False)
                continue

        # Here bookpath do not exists and we don't care for overwrite.
        try:
            book_write(book, bookpath)
        except NotADirectoryError as err:
            logger.warning("Book %s. Could not be written. %s", book.index, err)
            write_status.append(False)
        else:
            write_status.append(True)

    return write_status


def corpus_build_database(corpus_data: List[model.Book],
                          corpusdb_path: Union[str, Path]) -> database.TinyDB:
    """Returns the book database from a list of books"""

    corpusdb: database.TinyDB = TinyDB(corpusdb_path)

    logger.info("Building corpus %s with %d contents",
                corpusdb_path,
                len(corpus_data))

    corpusdb.insert_multiple(book.dict() for book in corpus_data)
    logger.info("Corpus completed with %d entries", len(corpusdb))

    return corpusdb


def corpus_load(rebuild: bool = False) -> database.TinyDB:
    """Load corpus. Download if needed. Returns the corpus database
    It loads the corpus from CORPUS_DATAPATH directory.

    Inputs:
        rebuild (Bool): Rebuild corpus.db even if it exists. Useful when
                        data is corrupted or was changed directly on file
                        (not so common).

    Outputs:
        corpus (database.TinyDB): Corpus as pandas dataframe or None
    """

    logger.info("Starting library version %s", constants.__version__)

    # Makes sure corpus if available
    corpus_datapath: Path = utils.get_corpus_datapath()
    corpus_url: str = utils.get_corpus_url()
    corpus_retrieve(corpus_datapath=corpus_datapath,
                    corpus_url=corpus_url)

    # Corpus database file
    corpusdb_filename: str = 'corpus.db'
    corpusdb_filepath: Path = corpus_datapath / corpusdb_filename

    # Build the corpus database
    corpusdb: database.TinyDB
    if not corpusdb_filepath.exists() or rebuild:
        # Corpus database do not exist (or rebuild). Lets build it.

        # Read corpus from book's dir
        booksdir: Path = utils.book_dir(corpus_datapath)
        books: List[model.Book] = corpus_read(booksdir)

        # Build corpus
        corpusdb = corpus_build_database(books, corpusdb_filepath)
    else:
        # Corpus database exists. Just load it.
        logger.info("Loading corpus dataset %s", corpusdb_filepath)
        corpusdb = TinyDB(corpusdb_filepath)

    return corpusdb


def corpus_insert_book(corpusdb: database.TinyDB, book: model.Book) -> List[int]:
    """Upsert (insert + update) book into corpusdb and write it to disk"""
    book_document: Dict = book.dict(exclude_defaults = True,
                                    exclude_none = True,
                                    exclude_unset = True)

    document_id: List[int] = corpusdb.upsert(book_document, Query().index == book.index)

    # Write book to bookdir
    booksdir: Path = utils.book_dir(utils.get_corpus_datapath())
    book_write(book, booksdir / book.index)

    return document_id


def corpus(corpusdb: database.TinyDB,
           query: Optional[queries.QueryInstance] = None) -> List[str]:
    """Return a corpus list from corpusdb and a query (optional)

    Input:
        corpusdb (db): A Corpus database handler
        query (query): A query request. If no query is given, returns the whole corpus.

    Output:
        corpus (list): A list where each element is a document from corpus.
    """
    result: List[table.Document]
    if query is None:
        result = corpusdb.all()
    else:
        result = corpusdb.search(query)

    local_corpus: List[str] = [ entry.get('contents', '') for entry in result ]

    return local_corpus


def metadata(corpusdb: database.TinyDB,
             query: Optional[queries.QueryInstance] = None) -> List[model.Book]:
    """Return a list of Book models given corpusdb and a query (optional).
    If no query is given, return all documents.

    Input:
        corpusdb (db): A corpus database handler
        query (query): A query request. If no query is given, return all documents.

    Output:
        List[model.Book]: A list of book metadata (without contents)
    """
    result: List[table.Document]
    if query is None:
        result = corpusdb.all()
    else:
        result = corpusdb.search(query)

    model_args: Dict[str, Any] = {'exclude': {'contents'},
                                  'exclude_defaults': True,
                                  'exclude_none': True,
                                  'exclude_unset': True}

    book: model.Book
    books: List[model.Book] = []
    for entry in result:
        book = model.Book(**entry)
        book_dict = book.dict(**model_args)
        book = model.Book(**book_dict)
        books.append(book)

    return books


def doc_id(corpusdb: database.TinyDB,
           query: Optional[queries.QueryInstance] = None) -> List[int]:
    """Return a list of document ids given a query

    Input:
        corpusdb (db): A Corpus database handler
        query (query): A query request. If no query is given, returns the whole corpus.

    Output:
        doc_ids (list): A list where each element is a document id from corpus.
    """
    result: List[table.Document]
    if query is None:
        result = corpusdb.all()
    else:
        result = corpusdb.search(query)

    doc_ids: List[int] = [ entry.doc_id for entry in result ]
    return doc_ids


if __name__ == '__main__':
    pass



"""Functions to read epub and convert to litcorpt Book model"""

from typing import Dict, List, Generator, Iterator
from pathlib import Path

import ebooklib
from ebooklib import epub

from bs4 import BeautifulSoup as bs

from . import model


def epub_metadata_field_get(epub_handler: ebooklib.epub.EpubBook,
                            field_name: str) -> List[str]:
    """Retrieve a DC epub metadata and filter it"""
    field_value = epub_handler.get_metadata('DC', field_name)
    values = [entry[0]
              for entry in field_value
              if entry[0] is not None]

    # Convert html to text in description
    if field_name == 'description':
        values = list(map(lambda x: bs(x, 'html.parser').text.strip(), values))

    return values


def epub_metadata(epub_handler: ebooklib.epub.EpubBook) -> Dict:
    """Return a dictionary with epub metadata"""
    metadata_fields = ['identifier', 'title', 'language', 'creator',
                       'description', 'subject',
                       # Often trash data, just ignore
                       # 'format',
                       # 'relation',
                       # 'source',
                       # 'type',
                       # 'contributor',
                       # 'date',
                       ]

    metadata: Dict = {}
    for field_name in metadata_fields:
        field_value = epub_metadata_field_get(epub_handler, field_name)
        if field_value:
            metadata[field_name] = field_value

    return metadata


def epub_metadata_filter_identifier(identifier: List[str]) -> List[str]:
    """Filter identifier field"""
    identifier_filtered = []

    invalid_matches = [ 'urn', 'uuid', 'AWP', 'awp', '{' ]

    for ident in identifier:
        if not 10 < len(ident) < 23:
            continue

        if any(match in ident for match in invalid_matches):
            continue

        identifier_filtered.append(ident)

    return identifier_filtered


def epub_metadata_filter_language(language: List[str]) -> List[str]:
    """Filter language field"""
    language_filtered = []
    for lang in language:
        if lang.lower() == 'pt-br':
            lang = 'pt_BR'
        language_filtered.append(lang)

    return language_filtered


def epub_metadata_filter_creator(creator: List[str]) -> List[Dict]:
    """Filter creator field"""
    creator_filtered: List[str] = []
    for creat in creator:
        if creat.count(',') == 1:
            # We have a SURNANE, Name
            creat = ' '.join(list(map(lambda x: x.strip(), creat.split(',')))[::-1])
            creator_filtered.append(creat.strip())
        elif creator.count(',') > 1:
            # We proably have a list of authors separated by 'commas'.
            creat_list = creat.split(',')
            for creat_slice in creat_list:
                creator_filtered.append(creat_slice.strip())
        else:
            # We have a Name Surname
            creator_filtered.append(creat.strip())

    creator_dict: List[Dict] = []
    for creat in creator_filtered:
        creat_dict = {'role': 'author'}

        name = creat.split()
        if len(name) == 1:
            creat_dict['lastname'] = name[0]
        elif len(name) > 1:
            creat_dict['lastname'] = name[-1]
            creat_dict['firstname'] = ' '.join(name[:-1])

        creator_dict.append(creat_dict)

    return creator_dict


def epub_metadata_to_book(metadata: Dict, index: str) -> model.Book:
    """Return Book metadata from ePub"""
    local_metadata = metadata.copy()

    local_metadata['index'] = index

    if 'identifier' in metadata and metadata['identifier']:
        local_metadata['identifier'] = epub_metadata_filter_identifier(local_metadata['identifier'])

    if not local_metadata['identifier']:
        del local_metadata['identifier']

    if 'language' in metadata and metadata['language']:
        local_metadata['language'] = epub_metadata_filter_language(local_metadata['language'])

    if 'creator' in metadata and metadata['creator']:
        local_metadata['creator'] = epub_metadata_filter_creator(local_metadata['creator'])

    book = model.Book(**local_metadata)

    return book


def epub_contents_text(epub_items: Iterator[ebooklib.epub.EpubHtml]) -> str:
    """Convert a epub content to text"""
    book_contents: List[str] = []
    book_contents = [bs(epub_item.get_body_content(), "html.parser").text
                     for epub_item in epub_items]

    contents: str = "".join(book_contents)
    return contents


def epub_filter_bodymatter(epub_items: Iterator[ebooklib.epub.EpubHtml],
                          ) -> Generator[ebooklib.epub.EpubHtml, None, None]:
    """Filter some epub content"""
    # Remove some useless epub content.
    for epub_item in epub_items:
        if "odin" in epub_item.get_name().lower():
            continue
        if "copyright" in epub_item.get_name().lower():
            continue
        if "toc" in epub_item.get_name().lower():
            continue
        if "also" in epub_item.get_name().lower():
            continue

        yield epub_item


def epub_filter_text(epub_items: Iterator[ebooklib.epub.EpubItem],
                    ) -> Generator[ebooklib.epub.EpubHtml, None, None]:
    """Filter out all non text content from epub"""
    for epub_item in epub_items:
        if epub_item.get_type() == ebooklib.ITEM_DOCUMENT:
            yield epub_item


def get_epub_items(epub_handler: ebooklib.epub.EpubBook,
                  ) -> Generator[ebooklib.epub.EpubItem, None, None]:
    """Generator for epub items in epub"""
    for epub_item in epub_handler.get_items():
        yield epub_item


def read(path: Path, index: str) -> model.Book:
    """Receive a epub file path and convert to Book Model indexed by index.
    Index will be used to uniquely identify the book in database, also as a directory
    name when writing to disk.
    """
    epub_handler: ebooklib.epub.EpubBook = epub.read_epub(path)
    epub_contents: Iterator[ebooklib.epub.EpubItem] = get_epub_items(epub_handler)
    epub_contents = epub_filter_text(epub_contents)
    epub_contents = epub_filter_bodymatter(epub_contents)
    epub_text: str = epub_contents_text(epub_contents)
    metadata: Dict = epub_metadata(epub_handler)
    metadata['contents'] = epub_text
    book = epub_metadata_to_book(metadata, index)

    return book
